# -*- coding: utf8 -*-
from __future__ import unicode_literals

import collections

import six

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimList,
    HprimMultiple,
    HprimString,
    HprimDate,
    HprimForeign,
)


class DmsSegmentH(HeaderSegment):
    sender_name = HprimString((5, 1))
    software_version = HprimString((5, 2))
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    infos = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class DmsSegmentP(Segment):
    practice_assign_id = HprimString(3)
    lab_assigned_patient = HprimString(4)
    patient_id_n3 = HprimString(5)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    date_of_birth = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    attending_physician_code = HprimString((14, 1))
    attending_physician_last_name = HprimString((14, 2))
    attending_physician_first_name = HprimString((14, 3))
    patient_height = HprimString((17, 1))
    patient_height_unit = HprimString((17, 2))
    patient_weight = HprimString((18, 1))
    patient_weight_unit = HprimString((18, 2))
    patient_diagnosis_code = HprimString(19)
    admission_date = HprimDate(24)

    orders = HprimForeign("O", "patient")


class DmsSegmentO(Segment):
    specimen_id = HprimString(3)
    universal_test_id = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        test_id=HprimString(5),
        # Empty: perform the test
        # RR: repeat the test
        rerun=HprimString(6),
        dilution=HprimString(7),
    ))
    # R: Routine
    # S: Stat
    priority = HprimString(6)
    collection_date = HprimDate(8)
    # N: New Order
    # A: Add Order
    # C: Cancel Order
    action_code = HprimString(12)
    # S: Serum
    # C: CSF
    # P: Plasma
    # U: Urine
    # W: Whole Blood
    # A: Amniotic Fluid
    # O: Other
    specimen_type = HprimString((16, 1))
    specimen_source = HprimString((16, 2))
    # O: Order
    report_type = HprimString(26, default="O")

    results = HprimForeign("R", "order")
    serotheque_infos = HprimForeign("S018", "header")


class DmsSegmentR(Segment):
    test_id = HprimMultiple(
        3,
        code_test=HprimString(4),
        dilution=HprimString(5),
        test_dms=HprimString(7),
        reagent_lot=HprimString(8),
        reagent_serial_no=HprimString(9),
        control_lot_no=HprimString(10),
        cod_parametre=HprimString(11),
        # F: Final result
        # P: Preliminary result
        result_type=HprimString(12),
    )
    value_quantitative = HprimString((4, 1))
    value_qualitative = HprimString((4, 2))
    flags = HprimList(7)
    # F: Final result
    # P: Preliminary result (Rerun)
    result_status = HprimString(9)
    validation_info = HprimMultiple(
        11,
        id_technician=HprimString(1),
        id_validator=HprimString(2),
        date_result=HprimDate(3),
    )
    date_result = HprimDate(13)
    instrument_code = HprimString((14, 1))
    instrument_num_serial = HprimString((14, 2))
    mean_previous_result = HprimString(15)
    image_file = HprimString(16)
    module = HprimString((17, 1))
    rack_id = HprimString((17, 2))
    sample_position = HprimString((17, 3))
    additional_information = HprimString((17, 4))

    infos = HprimForeign("M", "result")


class DmsSegmentM(Segment):
    source = HprimString(3)
    qc_code = HprimMultiple(
        4,
        qc_code=HprimString(1),
        instrument_code=HprimString(2),
        lot_number=HprimString(3),
        reagent_lot=HprimString(4),
        num_serial=HprimString(5),
    )
    test_name = HprimString(5)
    qc_result = HprimString(6)
    date = HprimDate((7, 1))
    unit = HprimString((7, 2))
    mean_theory = HprimString(8)
    sd_theory = HprimString(9)
    cv_theory = HprimString(10)
    # V: Valid
    # B: Bypassed
    # N: Not valid
    status_qc = HprimString(11)
    user = HprimString(12)
    calculated_qc_mean = HprimString(13)
    calculated_qc_sd = HprimString(14)
    calculated_qc_cv = HprimString(15)
    qc_delta = HprimString(16)
    num_serial_reagent = HprimString(17)
    flags = HprimString(18)
    units = HprimString(19)


class DmsSegmentC(Segment):
    comment_source = HprimString(3)
    comment_note_patient = HprimString((4, 1))
    comment_note_comment = HprimString((4, 2))
    comment_type = HprimString(5)


class DmsSegmentS018(Segment):
    comment_source = HprimString(3, default='L')
    message_code = HprimString((4, 1), default='S018')
    primary_sample_id = HprimString((4, 2))
    action_code = HprimString((4, 3), default='N')
    tubes = HprimList(5, basefield=HprimMultiple(
        0,
        secondary_sample_id=HprimString(1),
        label_type=HprimString(2),
        microliters=HprimString(3),
        srTest=HprimString(4),
        dept=HprimString(5),
        destination_code=HprimString(6),
        worklist=HprimString(7),
        test_name=HprimString(8),
        priority=HprimString(9),
        temperature=HprimString(10),
        specimen_type=HprimString(11),
        specimen_number=HprimString(12),
        alicot_indicator=HprimString(13),
        rack=HprimString(14),
        additional_information=HprimString(15),  # no de sérothèque
    ))
    comment_type = HprimString(6, default='S')


class DmsSegmentL(Segment):
    terminaison_code = HprimString(3)


class Driver(AstmMessageV2):
    SPEC = {
        "code": "dms",
        "label": "Dms Flexlab Inpeco"
    }
    SEGMENT_CLASS_MAP = {
        "H": DmsSegmentH,
        "P": DmsSegmentP,
        "M": DmsSegmentM,
        "O": DmsSegmentO,
        "R": DmsSegmentR,
        "C": DmsSegmentC,
        "S018": DmsSegmentS018,
        "L": DmsSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    OPTIONS = [
        AstmMessageV2.Option(
            "dms_qc_code_instrument",
            label="Codes analyseurs pour lesquels le lot de QC est dans le champ M.4.1 (séparés par ';')"),
        AstmMessageV2.Option(
            "dms_qc_acltop",
            label="Codes analyseurs des ACLTOP identifiés dans le champ M.4.2 (séparés par ';'"),
        AstmMessageV2.BooleanOption(
            "dms_envoyer_no_demandeur",
            label="Envoyer le numéro demandeur dans le champ P.3"),
        AstmMessageV2.BooleanOption(
            "dms_envoyer_no_patient_id",
            label="Envoyer le numéro IP Patient de MCA dans le champ P.4"),
        AstmMessageV2.Option(
            "analyseurs_hemato",
            label="Analyseurs d'hématologie (séparés par ';')"),
        AstmMessageV2.BooleanOption(
            "lxl_ref_qc_tronquee",
            label="Pour le LxL, ignorer le premier caractère '#' des lots de QC",),
        AstmMessageV2.BooleanOption(
            "dms_envoyer_no_serotheque",
            label="Envoyer le numéro sérotheque dans le champ 11.01.05.15"),
        AstmMessageV2.BooleanOption(
            "dms_ignore_analyseur_repasse",
            label="Ne pas envoyer l'analyseur pour les repasses dans le champ O.5.6"),
        AstmMessageV2.BooleanOption(
            "dms_dilution_en_alarme",
            label="Traiter le facteur du dilution comme alarme du résultat"),
        AstmMessageV2.IntegerOption(
            "dms_longueur_analyseur",
            label="La taille du champ `instrument` est limité à n caractères."),
    ]

    def __init__(
        self, dms_qc_code_instrument=None, dms_envoyer_no_demandeur=False,
        dms_envoyer_no_patient_id=False, analyseurs_hemato=None,
        lxl_ref_qc_tronquee=False, dms_envoyer_no_serotheque=False,
        dms_qc_acltop=None, dms_ignore_analyseur_repasse=False,
        dms_dilution_en_alarme=False, dms_longueur_analyseur=0,
        **kwargs
    ):
        self.qc_code_instrument = dms_qc_code_instrument.split(";") if dms_qc_code_instrument else []
        self.qc_acltop = dms_qc_acltop.split(";") if dms_qc_acltop else []
        self.envoyer_no_demandeur = dms_envoyer_no_demandeur
        self.envoyer_no_patient_id = dms_envoyer_no_patient_id
        self.lxl_ref_qc_tronquee = lxl_ref_qc_tronquee
        self.dms_envoyer_no_serotheque = dms_envoyer_no_serotheque
        self.dms_ignore_analyseur_repasse = dms_ignore_analyseur_repasse
        self.dms_dilution_en_alarme = dms_dilution_en_alarme
        self.instrument_length = dms_longueur_analyseur

        if analyseurs_hemato is not None:
            self.analyseurs_hemato = [
                e for e in analyseurs_hemato.split(";") if e
            ]
        else:
            self.analyseurs_hemato = []
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = DmsSegmentH(type_segment="H", **kwargs)
        message.end = DmsSegmentL(type_segment="L")
        return message

    def encode_worklist_enable_disable(self, worklist, context):
        message = self.create(date=context["now"])
        for sample in worklist.samples:
            for order in sample.orders:
                if order.action == "del":
                    source = "DISABLE"
                else:
                    source = "ENABLE"
                seg_m = message.infos.add(
                    source=source,
                    test_name=order.code,
                )
                seg_m.qc_code.qc_code = sample.sample_id.split(":", 1)[1]
        if message.infos:
            return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))

        enable_disable = {
            sample.sample_id.startswith("INSTRUMENT:")
            for sample in worklist.samples
        }
        if len(enable_disable) > 1:
            raise NotImplementedError("Mixed enable/disable and normal samples")

        if enable_disable == {True}:
            return self.encode_worklist_enable_disable(worklist, context)

        msg = self.create(
            date=context["now"]
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority", "rerun", "analyseur_repasse"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "C", "add": "A", "new": "N"},
        ):
            seg_p = msg.patients.add(
                last_name=patient.last_name,
                first_name=patient.first_name,
                mother_maiden_name=patient.maiden_name,
                date_of_birth=patient.birthdate,
                sex=patient.sex or "U",
            )
            if self.envoyer_no_patient_id:
                seg_p.lab_assigned_patient = patient.patient_id
            else:
                seg_p.lab_assigned_patient = patient.lab_patient_id
            if patient.height:
                seg_p.patient_height = patient.height
                seg_p.patient_height_unit = "cm"
            if patient.weight:
                seg_p.patient_weight = patient.weight
                seg_p.patient_weight_unit = "kg"
            for sample, action, priority, rerun, analyseur_repasse in patient.samples:
                seg_o = seg_p.orders.add(
                    specimen_id=sample.sample_id,
                    collection_date=sample.collection_date,
                    priority=priority,
                    action_code=action,
                    specimen_type=sample.specimen,
                )
                seg_p.admission_date = sample.admission_date
                if self.envoyer_no_demandeur:
                    seg_p.practice_assign_id = sample.request_id
                else:
                    seg_p.practice_assign_id = patient.patient_id
                for physician in sample.physicians:
                    seg_p.attending_physician_code = physician.code
                for order in sample.orders:
                    t = seg_o.universal_test_id.add()
                    t.code = order["code"]
                    t.dilution = order["dilution"]
                    if order.action == "add" and rerun:
                        if analyseur_repasse and not self.dms_ignore_analyseur_repasse:
                            t.rerun = analyseur_repasse
                        else:
                            t.rerun = "RR"
                if self.dms_envoyer_no_serotheque and action == 'N' and sample.info_serotheque:
                    segS018 = seg_o.serotheque_infos.add(
                        type_segment='C',
                        primary_sample_id=sample.sample_id,
                        action_code=action,
                    )
                    t = segS018.tubes.add()
                    t.secondary_sample_id = sample.sample_id
                    t.destination_code = sample.info_serotheque.destination
                    t.additional_information = sample.info_serotheque.numero

        if msg.patients:
            return msg

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for seg_m in message.infos:
            # Traitement des QC
            if seg_m.source == "QC":
                # Le flag D indique que le QC a été supprimé sur le DMS
                # Ne pas le traiter
                if seg_m.flags == "D":
                    continue
                if seg_m.qc_code.instrument_code in self.qc_code_instrument:
                    lot = seg_m.qc_code.qc_code
                elif seg_m.qc_code.instrument_code in self.qc_acltop:
                    lot = "{}-{}".format(seg_m.qc_code.qc_code, seg_m.qc_code.lot_number)
                else:
                    if (self.lxl_ref_qc_tronquee and
                            seg_m.qc_code.instrument_code.startswith("lxl") and
                            seg_m.qc_code.lot_number and
                            seg_m.qc_code.lot_number[0] == "#"):
                        # tronquer la référence de lot de QC du premier
                        # caractère lorsqu'il y a l'instrument lxl
                        lot = seg_m.qc_code.lot_number[1:]
                    else:
                        lot = seg_m.qc_code.lot_number

                sample = worklist.get_or_create(
                    sample_id=lot,
                    control_lot=lot,
                    type="control",
                )
                result = sample.results.add(
                    code=seg_m.test_name,
                    value=seg_m.qc_result,
                    date=seg_m.date,
                    instrument=seg_m.qc_code.instrument_code,
                )
                if seg_m.qc_code.reagent_lot:
                    result.lots.add(
                        type="reagent",
                        lot=seg_m.qc_code.reagent_lot.split(';')[0],  # éviter les doublons
                    )

        # Traitement des résultats patients
        # Dans une session de communication, il y a un segment P et autant de
        # segment O que de segment R pour un même échantillon
        liste_alarmes = collections.defaultdict(set)
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                date_rack = None
                for seg_r in seg_o.results:
                    sample = worklist.get_or_create(
                        sample_id=seg_o.specimen_id,
                        type="patient",
                    )
                    if seg_r.rack_id and seg_r.sample_position:
                        if date_rack is None or date_rack < seg_r.date_result:
                            date_rack = seg_r.date_result
                            sample.rack = "{}-{}".format(seg_r.rack_id, seg_r.sample_position)
                    commentaires = []
                    # uniquement les résultats avec le status F (Final)
                    if seg_r.result_status != "F":
                        continue
                    for seg_c in seg_r.commentaires:
                        if seg_c.comment_note_patient and seg_c.comment_note_patient != 'result':
                            commentaires.append(seg_c.comment_note_patient.strip())
                    # resultat quantitative et qualitative
                    for suffix, value in (
                        ("", seg_r.value_quantitative.strip()),
                        ("_I", seg_r.value_qualitative.strip()),
                    ):
                        alarmes = []
                        if seg_r.flags:
                            alarmes = seg_r.flags
                            # Traitement des flags alarmes uniquement pour les
                            # analyseurs d'hématologie
                            if sample.type == "patient" and seg_r.instrument_code in self.analyseurs_hemato:
                                for f in seg_r.flags:
                                    liste_alarmes[sample.sample_id].add(f)
                        # La dilution est dans le champ O.5.7 dans le cas d'un Liaison
                        # connecté à l'arrière de la DMS
                        test_order = seg_o.universal_test_id
                        for test in test_order:
                            if test.dilution:
                                dilution_order = test.dilution
                            else:
                                dilution_order = ""
                        # La dilution est dans le champ R.3.5 pour tous les
                        # autres analyseurs
                        dilution = seg_r.test_id.dilution or dilution_order
                        if self.dms_dilution_en_alarme:
                            alarmes.append("dilution: {}".format(dilution))
                        if self.instrument_length and isinstance(
                                self.instrument_length,
                                six.integer_types
                        ) and self.instrument_length > 0:
                            instrument = seg_r.instrument_code[:self.instrument_length]
                        else:
                            instrument = seg_r.instrument_code
                        result = sample.results.add(
                            code="{}{}".format(seg_r.test_id.code_test, suffix),
                            value=value,
                            flags=alarmes,
                            comments=commentaires,
                            date=seg_r.date_result,
                            instrument=instrument,
                            dilution=dilution,
                        )
                        if seg_r.validation_info.id_technician:
                            result.operator = seg_r.validation_info.id_technician
                        if seg_r.test_id.reagent_lot:
                            result.lots.add(
                                type="reagent",
                                lot=seg_r.test_id.reagent_lot.split(';')[0],  # éviter les doublons
                            )
            # Après le traitement du dernier segment R, ajout du résultat
            # de l'alarme par tube pour les analyseurs d'hématologie
            for sample_id, flags in liste_alarmes.items():
                sample = worklist.get_or_create(type="patient", sample_id=sample_id)
                sample.results.add(
                    code="ALARME",
                    value="",
                    comments=sorted(flags),
                )
        return worklist
